@@ -10,9 +10,11 @@
 
 typedef struct DispatcherObject{
     PyObject_HEAD
+    /* Holds borrowed references to PyCFunction objects */
     dispatcher_t *dispatcher;
     int can_compile;        /* Can auto compile */
-    PyCFunctionWithKeywords firstdef, fallbackdef;
+    /* Borrowed references */
+    PyObject *firstdef, *fallbackdef;
 } DispatcherObject;
 
 static int tc_int8;
@@ -27,7 +29,7 @@ static int tc_float32;
 static int tc_float64;
 static int tc_complex64;
 static int tc_complex128;
-static int BAISC_TYPECODES[12];
+static int BASIC_TYPECODES[12];
 
 static int tc_intp;
 
@@ -41,7 +43,7 @@ PyObject* init_types(PyObject *self, PyObject *args)
     #define UNWRAP_TYPE(S)                                              \
         if(!(tmpobj = PyDict_GetItemString(dict, #S))) return NULL;     \
         else {  tc_##S = PyLong_AsLong(tmpobj);                         \
-                BAISC_TYPECODES[index++] = tc_##S;  }
+                BASIC_TYPECODES[index++] = tc_##S;  }
 
     UNWRAP_TYPE(int8)
     UNWRAP_TYPE(int16)
@@ -76,8 +78,7 @@ PyObject* init_types(PyObject *self, PyObject *args)
     Py_RETURN_NONE;
 }
 
-static
-void
+static void
 Dispatcher_dealloc(DispatcherObject *self)
 {
     dispatcher_del(self->dispatcher);
@@ -108,16 +109,16 @@ static
 PyObject*
 Dispatcher_Insert(DispatcherObject *self, PyObject *args)
 {
-    PyObject *sigtup, *addrobj;
-    void *addr;
+    PyObject *sigtup, *cfunc;
     int i, sigsz;
     int *sig;
     int objectmode = 0;
 
-    if (!PyArg_ParseTuple(args, "OO|i", &sigtup, &addrobj, &objectmode)) {
+    if (!PyArg_ParseTuple(args, "OO!|i", &sigtup, &PyCFunction_Type,
+                          &cfunc, &objectmode)) {
         return NULL;
     }
-    addr = PyLong_AsVoidPtr(addrobj);
+
     sigsz = PySequence_Fast_GET_SIZE(sigtup);
     sig = malloc(sigsz * sizeof(int));
 
@@ -125,15 +126,17 @@ Dispatcher_Insert(DispatcherObject *self, PyObject *args)
         sig[i] = PyLong_AsLong(PySequence_Fast_GET_ITEM(sigtup, i));
     }
 
-    dispatcher_add_defn(self->dispatcher, sig, (void*)addr);
+    /* The reference to cfunc is borrowed; this only works because the
+       derived Python class also stores an (owned) reference to cfunc. */
+    dispatcher_add_defn(self->dispatcher, sig, (void*) cfunc);
 
     /* Add first definition */
     if (!self->firstdef) {
-        self->firstdef = (PyCFunctionWithKeywords)addr;
+        self->firstdef = cfunc;
     }
     /* Add pure python fallback */
     if (!self->fallbackdef && objectmode){
-        self->fallbackdef = (PyCFunctionWithKeywords)addr;
+        self->fallbackdef = cfunc;
     }
 
     free(sig);
@@ -142,7 +145,6 @@ Dispatcher_Insert(DispatcherObject *self, PyObject *args)
 }
 
 
-
 static
 PyObject*
 Dispatcher_DisableCompile(DispatcherObject *self, PyObject *args)
@@ -198,10 +200,11 @@ int typecode_fallback(DispatcherObject *dispatcher, PyObject *val) {
     }
 
     tmpcode = PyObject_GetAttrString(tmptype, "_code");
+    Py_DECREF(tmptype);
+    if (tmpcode == NULL)
+        return -1;
     typecode = PyLong_AsLong(tmpcode);
-
-    Py_XDECREF(tmpcode);
-    Py_XDECREF(tmptype);
+    Py_DECREF(tmpcode);
     return typecode;
 }
 
@@ -301,13 +304,16 @@ int typecode_arrayscalar(DispatcherObject *dispatcher, PyObject* aryscalar) {
     Py_DECREF(descr);
     if (typecode == -1)
         return typecode_fallback(dispatcher, aryscalar);
-    return BAISC_TYPECODES[typecode];
+    return BASIC_TYPECODES[typecode];
 }
 
 
 static
 int typecode(DispatcherObject *dispatcher, PyObject *val) {
     PyTypeObject *tyobj = val->ob_type;
+    /* This needs to be kept in sync with Dispatcher.typeof_pyval(),
+     * otherwise funny things may happen.
+     */
     if (tyobj == &PyInt_Type || tyobj == &PyLong_Type)
         return tc_intp;
     else if (tyobj == &PyFloat_Type)
@@ -343,6 +349,16 @@ void explain_ambiguous(PyObject *dispatcher, PyObject *args, PyObject *kws) {
     Py_XDECREF(callback);
 }
 
+/* A custom, fast, inlinable version of PyCFunction_Call() */
+static PyObject *
+call_cfunc(PyObject *cfunc, PyObject *args, PyObject *kws)
+{
+    PyCFunctionWithKeywords fn;
+    assert(PyCFunction_Check(cfunc));
+    assert(PyCFunction_GET_FLAGS(cfunc) == METH_VARARGS | METH_KEYWORDS);
+    fn = (PyCFunctionWithKeywords) PyCFunction_GET_FUNCTION(cfunc);
+    return fn(PyCFunction_GET_SELF(cfunc), args, kws);
+}
 
 static
 PyObject*
@@ -355,8 +371,7 @@ Dispatcher_call(DispatcherObject *self, PyObject *args, PyObject *kws)
     int prealloc[24];
     int matches;
     int old_can_compile;
-    PyCFunctionWithKeywords fn;
-    PyObject *cac;                  /* compile and call function */
+    PyObject *cfunc;
 
     /* Shortcut for single definition */
     /*if (!self->can_compile && 1 == dispatcher_count(self->dispatcher)){
@@ -377,26 +392,36 @@ Dispatcher_call(DispatcherObject *self, PyObject *args, PyObject *kws)
         if (tys[i] == -1) goto CLEANUP;
     }
 
-    fn = (PyCFunctionWithKeywords)dispatcher_resolve(self->dispatcher, tys,
-                                                     &matches);
+    /* We only allow unsafe conversions if compilation of new specializations
+       has been disabled. */
+    cfunc = dispatcher_resolve(self->dispatcher, tys, &matches,
+                               !self->can_compile);
     if (matches == 1) {
         /* Definition is found */
-        retval = fn(NULL, args, kws);
+        retval = call_cfunc(cfunc, args, kws);
     } else if (matches == 0) {
         /* No matching definition */
         if (self->can_compile) {
             /* Compile a new one */
-            cac = PyObject_GetAttrString((PyObject*)self, "_compile_and_call");
-            if (cac) {
-                old_can_compile = self->can_compile;
-                self->can_compile = 0;
-                retval = PyObject_Call(cac, args, kws);
-                self->can_compile = old_can_compile;
-                Py_DECREF(cac);
+            PyObject *cfa;
+            cfa = PyObject_GetAttrString((PyObject*)self, "_compile_for_args");
+            if (cfa == NULL)
+                goto CLEANUP;
+            old_can_compile = self->can_compile;
+            self->can_compile = 0;
+            /* NOTE: we call the compiled function ourselves instead of
+               letting the Python derived class do it.  This is for proper
+               behaviour of globals() in jitted functions (issue #476). */
+            cfunc = PyObject_Call(cfa, args, kws);
+            self->can_compile = old_can_compile;
+            Py_DECREF(cfa);
+            if (cfunc != NULL) {
+                retval = call_cfunc(cfunc, args, kws);
+                Py_DECREF(cfunc);
             }
         } else if (self->fallbackdef) {
             /* Have object fallback */
-            retval = self->fallbackdef(NULL, args, kws);
+            retval = call_cfunc(self->fallbackdef, args, kws);
         } else {
             /* Raise TypeError */
             PyErr_SetString(PyExc_TypeError, "No matching definition");
@@ -413,15 +438,8 @@ CLEANUP:
         free(tys);
 
     return retval;
-
 }
 
-
-static PyMemberDef Dispatcher_members[] = {
-    {NULL},
-};
-
-
 static PyMethodDef Dispatcher_methods[] = {
     { "_insert", (PyCFunction)Dispatcher_Insert, METH_VARARGS,
       "insert new definition"},
@@ -454,21 +472,21 @@ static PyTypeObject DispatcherType = {
     0,                         /*tp_as_sequence*/
     0,                         /*tp_as_mapping*/
     0,                         /*tp_hash */
-    (PyCFunctionWithKeywords)Dispatcher_call,           /*tp_call*/
+    (PyCFunctionWithKeywords)Dispatcher_call, /*tp_call*/
     0,                         /*tp_str*/
     0,                         /*tp_getattro*/
     0,                         /*tp_setattro*/
     0,                         /*tp_as_buffer*/
-    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,        /*tp_flags*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
     "Dispatcher object",       /* tp_doc */
-    0,		                   /* tp_traverse */
-    0,		                   /* tp_clear */
-    0,                         /*tp_richcompare */
-    0,		                   /* tp_weaklistoffset */
-    0,		                   /* tp_iter */
-    0,		                   /* tp_iternext */
+    0,                         /* tp_traverse */
+    0,                         /* tp_clear */
+    0,                         /* tp_richcompare */
+    0,                         /* tp_weaklistoffset */
+    0,                         /* tp_iter */
+    0,                         /* tp_iternext */
     Dispatcher_methods,        /* tp_methods */
-    Dispatcher_members,        /* tp_members */
+    0,                         /* tp_members */
     0,                         /* tp_getset */
     0,                         /* tp_base */
     0,                         /* tp_dict */
@@ -481,7 +499,6 @@ static PyTypeObject DispatcherType = {
 };
 
 
-
 static PyMethodDef ext_methods[] = {
 #define declmethod(func) { #func , ( PyCFunction )func , METH_VARARGS , NULL }
     declmethod(init_types),
@@ -490,7 +507,6 @@ static PyMethodDef ext_methods[] = {
 };
 
 
-
 MOD_INIT(_dispatcher) {
     PyObject *m;
     MOD_DEF(m, "_dispatcher", "No docs", ext_methods)
