@@ -30,7 +30,24 @@ class CFBlock(object):
         return iter(self.body)
 
 
-Loop = collections.namedtuple("Loop", ("entries", "exits", "header", "body"))
+class Loop(
+    collections.namedtuple("Loop", ("entries", "exits", "header", "body"))):
+    """
+    A control flow loop, as detected by a CFGraph object.
+    """
+
+    __slots__ = ()
+
+    # The loop header is enough to detect that two loops are really
+    # the same, assuming they belong to the same graph.
+    # (note: in practice, only one loop instance is created per graph
+    #  loop, so identity would be fine)
+
+    def __eq__(self, other):
+        return isinstance(other, Loop) and other.header == self.header
+
+    def __hash__(self):
+        return hash(self.header)
 
 
 class CFGraph(object):
@@ -68,7 +85,7 @@ class CFGraph(object):
         Yield (node, data) pairs representing the successors of node *src*.
         (*data* will be None if no data was specified when adding the edge)
         """
-        for dest in self._succs:
+        for dest in self._succs[src]:
             yield dest, self._edge_data[src, dest]
 
     def predecessors(self, dest):
@@ -76,7 +93,7 @@ class CFGraph(object):
         Yield (node, data) pairs representing the predecessors of node *dest*.
         (*data* will be None if no data was specified when adding the edge)
         """
-        for src in self._preds:
+        for src in self._preds[dest]:
             yield src, self._edge_data[src, dest]
 
     def set_entry_point(self, node):
@@ -97,6 +114,8 @@ class CFGraph(object):
         self._find_exit_points()
         self._find_dominators()
         self._find_back_edges()
+        self._find_topo_order()
+        self._find_descendents()
         self._find_loops()
         self._find_post_dominators()
 
@@ -119,6 +138,13 @@ class CFGraph(object):
         """
         return self._post_doms
 
+    def descendents(self, node):
+        """
+        Return the set of descendents of the given *node*, in topological
+        order (ignoring back edges).
+        """
+        return self._descs[node]
+
     def exit_points(self):
         """
         Return the computed set of exit nodes (may be empty).
@@ -160,6 +186,26 @@ class CFGraph(object):
         """
         return self._nodes
 
+    def topo_order(self):
+        """
+        Return the sequence of nodes in topological order (ignoring back
+        edges).
+        """
+        return self._topo_order
+
+    def topo_sort(self, nodes, reverse=False):
+        """
+        Iterate over the *nodes* in topological order (ignoring back edges).
+        The sort isn't guaranteed to be stable.
+        """
+        nodes = set(nodes)
+        it = self._topo_order
+        if reverse:
+            it = reversed(it)
+        for n in it:
+            if n in nodes:
+                yield n
+
     def dump(self, file=None):
         """
         Dump extensive debug information.
@@ -315,6 +361,34 @@ class CFGraph(object):
             back_edges.update((src, dest) for dest in back)
         self._back_edges = back_edges
 
+    def _find_topo_order(self):
+        succs = self._succs
+        back_edges = self._back_edges
+        post_order = []
+        seen = set()
+
+        def _dfs_rec(node):
+            if node not in seen:
+                seen.add(node)
+                for dest in succs[node]:
+                    if (node, dest) not in back_edges:
+                        _dfs_rec(dest)
+                post_order.append(node)
+
+        _dfs_rec(self._entry_point)
+        post_order.reverse()
+        self._topo_order = post_order
+
+    def _find_descendents(self):
+        descs = {}
+        for node in reversed(self._topo_order):
+            descs[node] = node_descs = set()
+            for succ in self._succs[node]:
+                if (node, succ) not in self._back_edges:
+                    node_descs.add(succ)
+                    node_descs.update(descs[succ])
+        self._descs = descs
+
     def _find_loops(self):
         """
         Find the loops defined by the graph's back edges.
@@ -449,7 +523,7 @@ class ControlFlowAnalysis(object):
         self.graph = graph
 
         # Fill incoming
-        for b in utils.dict_itervalues(self.blocks):
+        for b in utils.itervalues(self.blocks):
             for out, pops in b.outgoing_jumps.items():
                 self.blocks[out].incoming_jumps[b.offset] = pops
 
