@@ -1,8 +1,7 @@
 from __future__ import print_function, division, absolute_import
 
-from itertools import product
 
-from numba import types
+from numba import types, intrinsics
 from numba.utils import PYVERSION
 from numba.typing.templates import (AttributeTemplate, ConcreteTemplate,
                                     AbstractTemplate, builtin_global, builtin,
@@ -194,7 +193,8 @@ class BinOpFloorDiv(ConcreteTemplate):
 @builtin
 class BinOpPower(ConcreteTemplate):
     key = "**"
-    cases = [signature(types.float64, types.float64, op)
+    cases = [signature(types.intp, types.intp, types.intp)]
+    cases += [signature(types.float64, types.float64, op)
              for op in sorted(types.signed_domain)]
     cases += [signature(types.float64, types.float64, op)
              for op in sorted(types.unsigned_domain)]
@@ -204,6 +204,13 @@ class BinOpPower(ConcreteTemplate):
              for op in sorted(types.complex_domain)]
 
 
+class PowerBuiltin(BinOpPower):
+    key = pow
+    # TODO add 3 operand version
+
+builtin_global(pow, types.Function(PowerBuiltin))
+
+
 class BitwiseShiftOperation(ConcreteTemplate):
     cases = [signature(op, op, types.uint32)
              for op in sorted(types.signed_domain)]
@@ -411,6 +418,8 @@ class SetItemArray(AbstractTemplate):
         assert not kws
         ary, idx, val = args
         if isinstance(ary, types.Array):
+            if ary.const:
+                raise TypeError("Constant array")
             return signature(types.none, ary, normalize_index(idx), ary.dtype)
 
 
@@ -447,6 +456,15 @@ class ArrayAttribute(AttributeTemplate):
     def resolve_size(self, ary):
         return types.intp
 
+    def resolve_sum(self, ary):
+        return types.BoundFunction(Array_sum, ary)
+
+    def resolve_prod(self, ary):
+        return types.BoundFunction(Array_prod, ary)
+
+    def resolve_flat(self, ary):
+        return types.NumpyFlatType(ary)
+
     def generic_resolve(self, ary, attr):
         if isinstance(ary.dtype, types.Record):
             if attr in ary.dtype.fields:
@@ -454,16 +472,22 @@ class ArrayAttribute(AttributeTemplate):
                                    layout='A')
 
 
-class Array_flatten(AbstractTemplate):
-    key = "array.flatten"
+class Array_sum(AbstractTemplate):
+    key = "array.sum"
 
     def generic(self, args, kws):
         assert not args
         assert not kws
-        this = self.this
-        if this.layout == 'C':
-            resty = this.copy(ndim=1)
-            return signature(resty, recvr=this)
+        return signature(self.this.dtype, recvr=self.this)
+
+
+class Array_prod(AbstractTemplate):
+    key = "array.prod"
+
+    def generic(self, args, kws):
+        assert not args
+        assert not kws
+        return signature(self.this.dtype, recvr=self.this)
 
 
 @builtin
@@ -748,10 +772,17 @@ class Enumerate(AbstractTemplate):
 
     def generic(self, args, kws):
         assert not kws
-        [it] = args
+        it = args[0]
+        if len(args) > 1 and not args[1] in types.integer_domain:
+            raise TypeError("Only integers supported as start value in "
+                            "enumerate")
+        elif len(args) > 2:
+            #let python raise its own error
+            enumerate(*args)
+
         if isinstance(it, types.IterableType):
             enumerate_type = types.EnumerateType(it)
-            return signature(enumerate_type, it)
+            return signature(enumerate_type, *args)
 
 
 builtin_global(enumerate, types.Function(Enumerate))
@@ -769,3 +800,16 @@ class Zip(AbstractTemplate):
 
 
 builtin_global(zip, types.Function(Zip))
+
+
+@builtin
+class Intrinsic_array_ravel(AbstractTemplate):
+    key = intrinsics.array_ravel
+
+    def generic(self, args, kws):
+        assert not kws
+        [arr] = args
+        if arr.layout in 'CF' and arr.ndim >= 1:
+            return signature(arr.copy(ndim=1), arr)
+
+builtin_global(intrinsics.array_ravel, types.Function(Intrinsic_array_ravel))
